#! /usr/bin/perl -w

#####################################################################
#
#              NAME:               SPSK_setup_v2.2.pl
#              AUTHOR:             Jenny Power
#              LANGUAGE:           perl
#              STARTED WORK:       Jan12, 2011
#              VERSION DATE:       Jun21, 2011
#####################################################################
# v1.1 - outline with basic prompts 
# v1.2 - format checks for user inputs 
# v1.3 - runs the region file script, opens in DS9 and prompts user for manipulation
# v1.4 - Has user select a nearby offset star, outputs coords to cache
# v1.5 - Cleaned up, unnecessary lines removed
# v2.1 - Uses Newer version of Eric Hooper's code to create the fiber bundle overlay
# v2.2 - Colour coded!
# v2.3 - Fixed issue with cache formatting
# v2.4 - inserted more checks on user input
# v2.5 - Fixed formatting of output
# v3.0 - Fixed issue with offsets around RA of 0h
#      - outputs fits and jpg of central target
#      - fixed the rounding error with the rotational offset
#####################################################################

# The purpose of this script is to take a user step by step through the optimization of the 
# Sparsepak Fiver bundle placement.  It uses the script mk_ifu_reg_3.1.pl written by Eric Hooper
# which creates a region file of the fiber bundle, and helps the user to optimize the placement.
# The script mk_ifu_reg_3.1,pl creates a DS9 region file of the sparsepak fiber bundle, given
# a set of coordinates.

# This script takes an observer through optimizing the fiber bundle placement on a target, 
# listing the central fiber coordinates and any necessary rotator offsets. This code also 
# determines offsets to a nearby offset star.

# In order to run this script, the user must have XPA installed in their path.
#####################################################################
use Getopt::Long;
use POSIX;
use English; 
use Term::ANSIColor;
#####################STEP1###########################################
# Retrieve the necessary user input to run mk_ifu_reg_2.3.2.pl: Output file name, RA, and DEC

system "clear";

# Prompt user for the RA (lists formatting options)
grabRA:
print "\nPlease enter the RA of the target in J2000:\n";
$RA=<STDIN>;
chomp($RA);

#Varify that RA is in the format hh mm ss.ss or hh:mm:ss.ss or ##h##m##.##s
#convert all inputs to hh:mm:ss.ss format
# if not display message "incorrect format for RA" and reprompt the
# observer for RA.
# Format RA to hh:mm:ss.ss to make compatible with ds9

   if ($RA =~m/([+-]?\d\d) (\d\d) (\d+|\d+\.\d+|\d+\.|\.\d+)/ )
      {$RA = $1.":".$2.":".$3;
      goto RAdone;}
   if ($RA =~m/([+-]?\d\d):(\d\d):(\d\d)/)
      {goto RAdone;}
   if ($RA =~m/([+-]?\d\d)d(\d\d)m(\d+|\d+\.\d+|\d+\.|\.\d+)s/ )
      {$RA = $1.":".$2.":".$3;
      goto RAdone;}
   else {
       print colored ['white on_red'],  "Incorrect format for RA, please try again\n\n";
#       print "Incorrect format for RA, please try again\n\n";
       print "Please enter the DEC of the target in one of the following formats: 
                     * dd mm ss.ss  (separated by spaces).  
                     * dd:mm:ss.ss  (separated by colons). 
                     * ##h##m##.##s (separated by the characters 'h', 'm', and 's' where the  ## denote digits. \n\n";
      goto grabRA;}

RAdone:
system "clear";
;

# Prompt user for the DEC of the target

grabDEC:
print "\nPlease enter the DEC of the target in J2000:\n";
$DEC=<STDIN>;
chomp($DEC);

#Varify that DEC is in the format 'dd mm ss.ss' or dd:mm:ss.ss or '##d##m##.##s'
# if not display message "incorrect format for DEC" and reprompt the
# observer for DEC.
# Format DEC to dd:mm:ss.ss to make compatible with ds9

   if ($DEC =~m/([+-]?\d\d) (\d\d) (\d+|\d+\.\d+|\d+\.|\.\d+)/ )
      {$DEC = $1.":".$2.":".$3;
      goto DECdone;}
   if ($DEC =~m/([+-]?\d\d):(\d\d):(\d\d)/)
      {
      goto DECdone;}
   if ($DEC =~m/([+-]?\d\d)d(\d\d)m(\d+|\d+\.\d+|\d+\.|\.\d+)s/ )
      {$DEC = $1.":".$2.":".$3;
      goto DECdone;}
   else {
       print colored ['white on_red'], "Incorrect format for DEC, please try again\n\n";
       print "Please enter the DEC of the target in one of the following formats: 
                     * dd mm ss.ss  (separated by spaces).  
                     * dd:mm:ss.ss  (separated by colons). 
                     * ##d##m##.##s (separated by the characters 'd', 'm', and 's' where the  ## denote digits. \n\n";
      goto grabDEC;}

DECdone:
system "clear";


#prompt the user for the name of object
   print"\n Please enter the name of the object:\n";
   $name = <STDIN>;
   chomp ($name);

system "clear";

if ($name eq "") {$name = 'a';
   print colored ['white on_red'], "Since no object name was given, a default name of \"a\" was assigned";}

# Make sure the object name has no spaces in it, and remove spaces for file names.
   $object = $name;
   $object =~ s/\s+//g;


# The region file containing the unaltered fiber bundle will be named: $object.reg
   $regionfile1 = "${object}.reg";




#####################################################################


##################STEP2##############################################
#Run the perl script mk_ifu_reg_2.3.2.pl to create region file of
#fiber bundle
#Format: 
#    mk_ifu_reg_2.3.2.pl sparsepak_astrometry_table.txt <regionfile name> --RA <RA value> --Dec <DEC value>

   `perl mk_ifu_reg_3.1.pl sparsepak_astrometry_table.txt $regionfile1 --RA  $RA --Dec  $DEC`;

   print "A region file for the specified target containing the sparsepak fiber overlay with no manual alterations has been generated: "; 
   print colored ['white on_blue'], "$regionfile1 \n"; 



######################################################################



###################STEP3#############################################
#Open image of field in DS9 and overlay the region file containing the fiber bundle

#Prompt user for image name of image containing field. 

#Ask if user has their own image of the field or would like to see an SAO image of thier field.
   print "\nDo you have an image of your field? (y or n)  If not an SAO image of your field will be used.\n";
OwnImage:
   $own = <STDIN>;
   chomp $own;

# if the user has their own image, open that image in DS9 and overlay the region file, else use a SAO image
   if ($own eq "y" || $own eq "Y" || $own eq "Yes" || $own eq "yes"  || $own eq "YES") {
      print "\nPlease enter the full name (including extension) of the image of your field:\n";
      $fieldimage = <STDIN>;
      chomp $fieldimage;
      system "ds9 $fieldimage -region load $regionfile1 -regions select all -regions system wcs -regions getinfo &";
 }


#If the observer does not have their own fits image, then load an SAO image of their field.
   elsif($own eq "n" || $own eq "N" || $own eq "No" || $own eq "NO") {
     #display a ds9 image of the field centered at the coordinates provided earlier 
     #(colour inverted)
      system "ds9 -dsssao size 12 12 arcmin -dsssao coord $RA $DEC sexagesimal -zoom 3 -cmap invert yes -regions load $regionfile1 -regions select all -regions system wcs -regions getinfo  &";
 }

else {
    print "\nPlease enter y or n.\n";
    goto OwnImage;}

system "clear";

# Instruct the observer how to manipulate the fiber bundle.  

print "\nTanslate and rotate the fiber bundle to the position you would like over your target image.  
\n TRANSLATION:
Select the representation of the fiber bundle by clicking on the center of one of the circles (four tiny squares should appear around the fiber bundle indicating it has been selected) and drag the fiber bundle to the desired position on the image.
\n ROTATION:
A window labled \"Composite\" should open when your DS9 image is opened (If not open it through \"Regions\"->\"Get Information\").  Record the initial angle if non-zero.  Then change the rotation to the desired orientation by adjusting the Angle.\n";

PLACED:
   print "\nWhen you are satisfied with the fiber placement, hit return\n";
   $happy = <STDIN>;
   chomp $happy;


#Incase of accidental input
   if ($happy ne ""){
     goto PLACED;
   }


system "clear";
print "The coordinates of the central fiber will be written to file: ";
print colored ['white on_blue'], "${object}_cache.txt\n\n";

# If happy with fiber placement, save everything and extract the central fiber coordinates
   if ($happy eq ""){

# Save the optimized fiber bundle and save a fits image of the final set up for a finding chart
      $regionfile_final = "${object}_final.reg";


#If happy with the fiber bundle placement, extract relevant information (final RA, DEC, Angle)
#Save the final regionfile containing the fiber bundle placement
#The coordinates of the central fiber will be extracted from this region file.
      system "xpaset -p ds9 regions select all";
      system "xpaset -p ds9 regions skyformat sexagesimal";
      system "xpaset -p ds9 regions save $regionfile_final";

      $filein = "${object}_final.reg";
      open FILE, "<$filein" or die "Can't open $filein: $!\n";

      while($line = <FILE>) {
    if($line =~ m/composite\(\s*(\S+)\s*,\s*(\S+)\s*,\s*(\S+)\s*\).*/)
    {
	$RA_Final = "$1";
	$DEC_Final = "$2"; 
        $Angle = sprintf("%02.f","$3");
	last;
    }
} 
}
if ($Angle > 180) {
  $Angle_Final = $Angle - 360;}
if ($Angle <= 180){
  $Angle_Final = $Angle;}

print "Central fiber position: ";
print colored ['blue'], "RA $RA_Final\tDEC $DEC_Final\tRot = $Angle_Final\n";

$RAdecimal = SEXAGESIMALtoDEGRA($RA_Final);
#print "$RAdecimal\n";
$DECdecimal = SEXAGESIMALtoDEGDEC($DEC_Final);
#print "$DECdecimal\n";
   

system "xpaset -p ds9 quit";
######################################################################


##################STEP4###############################################
#Open a second ds9 image of the field so that the observer may select and offset star

#Does the observer want to have an offset star.

#    print "\nDo you want to select an offset star?  This is recommended to help refine telescope pointing and to help ensure the telescope is pointed optimally.\n";
#    $okstar = <STDIN>;
#    chomp $okstar;

# print colour ['yellow on_magenta'], '$okstar';
# #Incase of accidental input
#    if ($okstar  ne "y" || $okstar ne "Y" || $okstar ne "yes" || $okstar ne "Yes" || $okstar ne "YES"){
#       $fileout = "${object}_cache.txt";
#       open OUTPUT, ">$fileout" or die "Can't open $fileout: $!\n";
#       print OUTPUT "$object\t\t$RA_Final\t$DEC_Final\t2000.0\t\tRot: $Angle_Final\n";
#       close OUTPUT;
#      goto NOSTAR;
#    }


#In order to select the best offset stars, filter the possible offset stars based on the observed wavelength region.

print "\nTime to select an appropriate offset star.  This star will be used to ensure the telescope is pointed in the optimal position and to check pointing.\n";

 goLambda:
#Have the observer enter their central wavelength
print "\nPlease enter your (approximate) central wavelength in Angstoms:\n";
$lambda = <STDIN>;
chomp $lambda;


system "clear";
# If blue filter by Bmag, if red filter by Rmag
if ($lambda > 1000 && $lambda <= 5000) {
   #lower magnitude limit
   $brightest = '\$Bmag>8';
   #upper magnitude limit
   $faintest = '\$Bmag<16';
   goto lambdadone;}
if ($lambda > 5000 && $lambda < 10001){ 
   #lower magnitude limit
   $brightest = '\$Rmag>8';
   #upper magnitude limit
   $faintest = '\$Rmag<16';
   goto lambdadone;}
else {
# If the observer enters a wavelength that is not within the typical range, prompt them to retry
   print colored ['white on_red'], "Invalid wavelength, please try again. \n";
   goto goLambda;}

lambdadone:

system "clear";

#Open a ds9 window with the stars within the magnitude range circled.
system "ds9 -dsssao size 12 12 arcmin -dsssao coord $RA_Final $DEC_Final sexagesimal -zoom 3 -cmap invert yes -catalog ua2 -catalog size 12 12 arcmin -catalog filter \"$brightest && $faintest\" -mode catalog &";

#Prompt the user to select a suitable offset star from the open ds9 window

print "\nSelect a suitable offset star from the ds9 image of the field by clicking on the star in the image.  When clicked the information of the selected star should appear highlighted in the \"Catalog Tool\" window.\n";


#Get the user to input the RAJ2000 and DEC2000 of the offset star. These inputs will be in decimal form
print "\nPlease enter the RAJ2000 value (in degrees as listed) of the selected offset star:\n";
RAJ:
$offsetRAJ2000 = <STDIN>;
chomp $offsetRAJ2000;
if ($offsetRAJ2000 =~m/([+-]?\d+|\d+\.\d+|\d+\.|\.\d+)/ )  {
    # Convert the RA from degrees to sexagesimal
$offsetRAS = DEGtoSEXAGESIMALRA($offsetRAJ2000);}
  else {
    print colored ['white on_red'], "Incorrect format.";
    print "  Please enter decimal form of RAJ2000 as listed in the catalog tool.\n";
goto RAJ;
}



print "Please enter the DEJ2000 value of the selected offset star:\n";
DECJ:
$offsetDEJ2000 = <STDIN>;
chomp $offsetDEJ2000;
if ($offsetDEJ2000 =~m/([+-]?\d+|\d+\.\d+|\d+\.|\.\d+)/ ) {
    # Convert the RA from degrees to sexagesimal
$offsetDECS = DEGtoSEXAGESIMALDEC($offsetDEJ2000);}
  else {
    print colored ['white on_red'], "Incorrect format.";
    print "  Please enter decimal form of DEJ2000 as listed in the catalog tool.\n";
  goto DECJ;
  }


system "clear";
#########################################################################


##################STEP5###############################################
#Create a file with the final information, formatted for a copy paste into a cache

# The coordinates of the central fiber are given above as RA_FINAL, DEC_FINAL, and Angle_final

#Compare the offset star position to the central fiber position and covert to an offset in arc seconds.


#Find offset from central fiber to offset star, convert to arcseconds

$offsetRA = sprintf("%.1f",DEGtoARCSECONDS($offsetRAJ2000 - $RAdecimal)) ;
$offsetDEC = sprintf("%.1f",DEGtoARCSECONDS($offsetDEJ2000 -$DECdecimal));

print "offsets: ";
print colored ['blue'], "E $offsetRA\"\tN $offsetDEC\"\n";
#########################################################################

##############STEP6######################################################
#Write to file the central fiber position in a file that is compatible with Cache Submission

$name2 = sprintf '%-32s', "$name";
$checkstar = sprintf ('%-32s', "Check Star");

#Write to file the central fiber position
      $fileout = "${object}_cache.txt";
      open OUTPUT, ">$fileout" or die "Can't open $fileout: $!\n";
      $hname = sprintf('%-32s', "Name");
      $hRA = sprintf('%-16s', "RA");
      $hDec = sprintf('%-16s', "DEC");
      print OUTPUT "$hname\t$hRA\t$hDec\tEpoch\t\tComments\n";

      print  OUTPUT "$name2\t$RA_Final\t$DEC_Final\t2000.0\t\tRot: $Angle_Final Offsets: N $offsetDEC, E $offsetRA\n$checkstar\t$offsetRAS\t$offsetDECS\t2000.0\n";
      close OUTPUT;

system "xpaset -p ds9 quit";

NOSTAR:

#Save a copy of the DS9 region to be used as a finders chart
      $findingchartfits = "${object}_final_finders.fits";
      $findingchartjpg = "${object}_final_finders.jpg";

print "A file containing the coordinates of the central fiber, rotator angle, and the offset in arcseconds to a nearby checkstar (if one has been selected), and the coordinates of that nearby check star has been written: ";
print colored ['white on_blue'], "${object}_cache.txt";
print "\nAs well, a region file with the final fiber placement (";
print colored ['white on_blue'],"${object}_final.reg";
print") and a finders chart of the area (";
print colored ['white on_blue'],"${object}_final_finders.fits";
print " and ";
print colored ['white on_blue'],"${object}_final_finders.jpg";
print") have been written.\n\n\n";


#create a jpg and fits of the area as a finders chart
system "ds9 -dsssao size 12 12 arcmin -dsssao coord $RA_Final $DEC_Final sexagesimal -zoom 3 -cmap invert yes -regions load $regionfile_final &";

print "A DS9 image of your final field setup should appear.  Once the image has loaded, hit return to save a finder chart of the area:\n";
<STDIN>;

system "xpaset -p ds9 saveimage jpeg ${object}_final_finders.jpg";

system "xpaset -p ds9 saveimage fits $findingchartfits";



##########################################################################
##########################################################################
#### Some useful subroutines #############################################
##########################################################################

#1) Convert dec given in degrees to sexagesimal
sub DEGtoSEXAGESIMALDEC{
    my $DECd = 0; my $DECm = 0; my $DECs = 0;
    if ($_[0] < 0){ 
	$DECd = sprintf("%02d",ceil($_[0]));
	$DECm = sprintf("%02d",abs(ceil(($_[0]-$DECd)*60)));
	$DECs = sprintf("%05.02f",(((abs($_[0]-$DECd)*60)-$DECm)*60)); }
    else {
	$DECd = sprintf("+%02d",floor($_[0]));
	$DECm = sprintf("%02d",floor(($_[0]-$DECd)*60));
	$DECs = sprintf("%05.02f",(((($_[0]-$DECd)*60)-$DECm)*60)); }
    return  "$DECd $DECm $DECs";
}

#print DEGtoSEXAGESIMALDEC('1.5083333');
#print "\n";

#2) Convert RA given in degrees to sexigesimal
sub DEGtoSEXAGESIMALRA {
    my $RAh = 0; my $RAm = 0; my $RAs = 0;
    $RAh = sprintf("%02d",floor($_[0]/15));
    $RAm = sprintf("%02d",floor(($_[0]/15-$RAh)*60));
    $RAs = sprintf("%05.02f",(($_[0]/15-$RAh)*60-$RAm)*60);
    return  "$RAh $RAm $RAs";
}
#print DEGtoSEXAGESIMALRA('90.0');
#print "\n";

#3) Convert RA given in sexagesimal to degrees
sub SEXAGESIMALtoDEGRA{
   if ($_[0] =~m/([+-]?\d\d) (\d\d) (\d+|\d+\.\d+|\d+\.|\.\d+)/ )
      {return   $1 * 15 + $2 * 0.25 + $3 / 240;}
   if ($_[0] =~m/([+-]?\d\d):(\d\d):(\d\d)/)
      {return   $1 * 15 + $2 * 0.25 + $3 / 240;}
   if ($_[0] =~m/([+-]?\d\d)h(\d\d)m(\d+|\d+\.\d+|\d+\.|\.\d+)s/ )
      {return   $1 * 15 + $2 * 0.25 + $3 / 240;}
   else {die "Incorrect format.\n";}
    }

#$RA = '06 00 00';
#print SEXAGESIMALtoDEGRA($RA);
#print "\n";

#4) Convert DEC given in sexagesimal to degrees
sub SEXAGESIMALtoDEGDEC{
  if ($_[0] =~m/([+-]?\d\d) (\d\d) (\d+|\d+\.\d+|\d+\.|\.\d+)/ ){
     if ($1 < 0){return  -1*(abs($1)  + $2/60 + $3/3600);}
     if ($1 >= 0) {return  abs($1)  + $2/60 + $3/3600;}}
  if ($_[0] =~m/([+-]?\d\d):(\d\d):(\d\d)/){
     if ($1 < 0){return  -1*(abs($1)  + $2/60 + $3/3600);}
     if ($1 >= 0) {return  abs($1)  + $2/60 + $3/3600;}}
  if ($_[0] =~m/([+-]?\d\d)d(\d\d)m(\d+|\d+\.\d+|\d+\.|\.\d+)s/ ){
     if ($1 < 0){return  -1*(abs($1)  + $2/60 + $3/3600);}
     if ($1 >= 0) {return  abs($1)  + $2/60 + $3/3600;}}
   else {die "Incorrect format";}
 }
#$DEC = '-01 30 30';
#print SEXAGESIMALtoDEGDEC($DEC);
#print "\n";


sub DEGtoARCSECONDS {
   chomp $_[0];
   $temp = $_[0]*3600;
   if ($temp > '648000'){
     return $temp-'1296000';}
   if ($temp <= '648000'){
     return $temp;}
   else {die "Unknown input.";}
 }
