#! /usr/bin/perl

#####################################################################
#
#              NAME:               mk_ifu_reg_3.1.pl
#              AUTHOR:             Eric Hooper, UW-Madison
#              LANGUAGE:           perl
#              STARTED WORK:       autumn, 2007
#              ORIGINAL FILE NAME: mk_sparsepak_reg.pl
#              LAST MODIFICATION:  28 FEB 2011

# This script takes a table giving the relative positions and numeric
# descriptions of IFU elements, plus a sky position, and generates a
# ds9 region file which represents the IFU centered on the given
# coordinates.  The user then displays an image of the target field in
# ds9 and loads the region file generated by this script.  With the
# individual regions (e.g., a circle for each fiber or a rectangle for
# each micromirror) grouped together, the user can easily translate
# and rotate the IFUelement positions with respect to the field.  Once
# the IFU is aligned as desired, the rotation angle can be read from
# ds9, which can be easily converted to a position angle: PA =
# (initial rotation angle) - (final position angle).

# This script is routinely used with Sparsepak on WIYN.  It will 
# soon be employed with Hexpak on WIYN and may be adopted for  
# IFUs on other telescopes. 


# ------------------------------------------------------


# Command line FILEs, in the following *order* (these must be first).

# IFU element      IMPORTANT:  most users will not have to worry about 
# file (input).    the contents of this file.  An IFU element input 
#                  file will be provided.  The following description 
#                  of the IFU input file is important only for developers, 
#                  those seeking to use this script for a new IFU, 
#                  or the rabidly curious.  
#
#                  This file contains the relative positions of and other  
#                  information for each fiber, one line per fiber. The 
#                  columns are separated in this file by a delimeter string 
#                  that can be set from the command line with the switch 
#                  --input_delimeter (see below). 
#
#                  Several of the columns give angles on the sky
#                  relating to the position of the elements.  The
#                  units of these values are auto-detected, but if no
#                  valid unit is apparent, then the defaults set by
#                  --ifu_rel_pos_unit_ra and --ifu_rel_pos_unit_dec
#                  are used.  For both RA and DEC, the valid units are
#                  arcsec (any string beginning with 's'; any string
#                  beginning with "arcs" or the double quote: ");
#                  arcmin (any string beginning with 'm'; any string
#                  beginning with "arcm" or the double quote: ");
#                  degree (any string beginning with 'd') and radian
#                  (any string beginning with 'r').  
#                  *Important*: These values are angles on the sky,
#                  *not* coordinate differences (the distinction is #
#                  important in the RA direction).  Do *not* give the
#                  relative offsets # in this file in units of hours,
#                  minutes of time, # or seconds of time (hms-type
#                  units).
#                  *Important*: Only the position parameters have
#                  default units.  You have to include the units for
#                  the element size parameters as part of their
#                  entries in the IFU element file.
#                   
#                  The columns are:
#               1) fiber id, which can be an arbitrary string (without
#                  spaces) at this point, except that it cannot have the 
#                  comment character (#) as the first non-blank character. 
# 
#               2) Type of region element, such as circle, etc.
#                  See the ds9 documentation about specific region types. 
#
#          3 -> N) The next 2 or more columns list coordinates that
#                  define the position and/or configuration of the
#                  element.  For many elements, this is just the RA
#                  and DEC offsets, in that order, of each fiber.
#                  This is the case for common shapes, such as circle,
#                  ellipse, box, etc.  However, some elements need
#                  more than two position arguments, for example line,
#                  ruler, polygon, etc.
#
#        N+1 -> M) The next set of columns list parameters related to 
#                  the size and shape of the element, such as radii or 
#                  rotation angles.  For example, the only value for 
#                  the region type "circle" in this category of entries 
#                  is its radius.  *A unit string recognizable to ds9 
#                  should be post-pended to each of these values.*


# Region file for  After displaying the image of the target in ds9, 
# use with ds9     load this region file to see the representation of 
# (output).        the fiber bundle overlayed on the image.  Group 
#                  these regions together in ds9 to enable translation 
#                  and rotation of all of the fibers as a group. 


# ------------------------------------------------------


# Command line SWITCHES (follow the two names of files described
# above).  In typical applications, only --RA and --DEC need to be
# set.

# --RA             The right ascension of the central IFU element.  
#                  (For Sparsepak, this is typically fiber # 52,
#                  which has offsets (0,0). )
#
#                  The RA coordinate can be in one of the following
#                  formats.  There is no need to specify the format,
#                  for it will be auto detected.

#                  ** Decimal degrees 
#                     (positive value; digits; decimal point; more digits)
#                  ** Sexagesimal hours with any *one* (don't mix!) 
#                     of the following formats:
#                     * hh mm ss.ss  (separated by spaces).  
#                                    Need to put quotes around 
#                                    this string on the command line!!!
#                     * hh:mm:ss.ss  (separated by colons). 
#                     * ##h##m##.##s (separated by the characters 
#                       'h', 'm', and 's',where the  ## denote digits.
#                       Note that it is up to you whether to have
#                       spaces around the 'h', 'm', and 's' characters, 
#                       but if you do then don't forget to quote
#                       the entire string.)

# --DEC            The declination of the central IFU element.  The available
#                  formats are the same as for RA_center, except that
#                  the unit of the first entry in sexagesimal format
#                  is degrees; and if one uses characters to separate
#                  the elements in sexagesimal notation, the character
#                  for degrees is 'd'.  As an example of the latter,
#                  the appropriate string would be: ##d ##m ##.##s
#                  where ## denote digits and spaces around the 'd',
#                  'm', and 's' characters are optional, but if you
#                  include spaces don't forget to quote the entire
#                  string.

# --mk_composite   If this is set to true (non-zero; set to proper
#                  boolean later), then the region file produced by
#                  this script will have the composite flag set, which
#                  saves a step later in ds9 for most intended uses of
#                  this script.

# --color          The color of the fiber outlines in the ds9 display.
#                  This can be altered interactively in ds9.  Currently
#                  this parameter is utilized only if --mk_composite
#                  is set to true.

# --line_width     The width of the lines delineating the fiber outlines
#                  in the ds9 display.  This can be altered
#                  interactively in ds9.  Currently only implemented
#                  if --mk_composite is set to true.

# --ds9_delimeter  The delimeter which separates the sections of the
#                  ds9 regions file.  This must be a character
#                  recognized for this purpose by ds9.  It is unlikely
#                  this will need to be changed unless ds9 changes in
#                  a way that the default delimeter no longer works.

# #### The following refer to the input file and likely won't need to 
#      be changed during routine operation. 

# --input_delimeter      Alerts the code to what to expect for the 
#                        column separators in the input file.  This is
#                        currently set to ';' by default rather than
#                        the more common ',' in case we later decide
#                        to include any text elements in the region
#                        file.  A comma is more likely to show up in
#                        text than a semi-colon.

# --inp_fiber_id_pos     Column position (starting from 0) of fiber id 
#                        in the input file.

# --inp_elem_type_pos    The location of the name of the region element 
#                        (e.g., circle).

# --coord_elem_offset    Start reading coordinates for the element at 
#                        this array position (starting from 0).  For 
#                        example, if there's an id string and an 
#                        element type in the first two columns, and the 
#                        coordinate values begin in the third column, 
#                        this number should be 2. 

# --ifu_rel_pos_unit_ra  This is the *default* unit assumed for any 
#                        RA relative position values in the IFU 
#                        element input file.  Any proper indication 
#                        of units in the file itself will override 
#                        this default. 

# --ifu_rel_pos_unit_dec Same as above, but for DEC.  




# ---------------    Current Defaults ------------------------------

#                       This is just a convenient listing;
#                       the actual setting of the default values is in
#                       the code below.  Note: don't diddle most of
#                       these.  See discussion in description of
#                       parameters above.  For normal operation you
#                       need only set the RA and DEC.

#                       RA                                   0.0
#                       DEC                                  0.0
#                       mk_composite                         1   (true)
#                       color                                "blue"
#                       line_width                           3
#                       ds9_delimeter                        ','  
#                                                            (the delimeter
#                                                             is a comma)
#                       input_delimeter                      ';'
#                       inp_fiber_id_pos                     0
#                       inp_elem_type_pos                    1
#                       coord_elem_offset                    2
#                       ifu_rel_pos_unit_ra                  "arcsec"
#                       ifu_rel_pos_unit_dec                 "arcsec"


# Revision history:

# See the end of the file.


#
#
#####################################################################



###########################    MAIN    ##############################



use Getopt::Long;
use POSIX;
use English;


# ..............Parse the command line. .............................


# shift works on @ARGV by default in main, @_ in subroutines.

$ifu_element_position_file     = shift;
$output_region_file            = shift; 


# Make sure the file names are NOT flags, as would happen if a user
# forgot to begin with both file names.
die ("ERROR:  \$ifu_element_position_file = $ifu_element_position_file\n")  if ($ifu_element_position_file =~ /\-\-/);
die ("ERROR:  \$output_region_file = $output_region_file\n") if ($output_region_file =~ /\-\-/);


# Get the options.

%optctl = ();   # Line breaks in the &GetOptions correspond to parameter
                # categories, as listed in the documentation at the
                # top of the script.
&GetOptions (\%optctl, "RA=s","DEC=s","ds9_delimeter=s","mk_composite=i",
             "color=s","line_width=i",
             "input_delimeter=s","inp_fiber_id_pos=i","inp_elem_type_pos=i",
             "coord_elem_offset=i","ifu_rel_pos_unit_ra=s",
             "ifu_rel_pos_unit_dec=s"
             );


# Set the defaults for the potential command-line inputs. 


$optctl{RA}                  =   0.0               unless(defined $optctl{RA});
$optctl{DEC}                 =   0.0               unless(defined $optctl{DEC});
$optctl{ds9_delimeter}       =   ','               unless(defined $optctl{ds9_delimeter});
$optctl{mk_composite}        =   1                 unless(defined $optctl{mk_composite});
$optctl{color}               =   "blue"            unless(defined $optctl{color});
$optctl{line_width}          =   3                 unless(defined $optctl{line_width});
$optctl{input_delimeter}     =   ';'               unless(defined $optctl{input_delimeter});
$optctl{inp_fiber_id_pos}    =   0                 unless(defined $optctl{inp_fiber_id_pos});
$optctl{inp_elem_type_pos}   =   1                 unless(defined $optctl{inp_elem_type_pos});
$optctl{coord_elem_offset}   =   2                 unless(defined $optctl{coord_elem_offset});
$optctl{ifu_rel_pos_unit_ra} =   "arcsec"          unless(defined $optctl{ifu_rel_pos_unit_ra});
$optctl{ifu_rel_pos_unit_dec}=   "arcsec"          unless(defined $optctl{ifu_rel_pos_unit_dec});


#  ............End of parsing the command line. ...............




#  ........... Set internal defaults.                    ...............



# Set variables that likely won't need to be modified from the command
# line.  If later we find that they do, move them to command-line
# switches, or if it's too complicated, read in another file with
# these definitions.

# Give the positions in the input array of the coordinate values (they
# are input relative to the center of the IFU bundle).  We need to
# separate out those which are in the declination axis, since those
# will have the user-specified $optctl{DEC} added to these offsets.
# Similarly, we will have to separate out those along the right
# ascension axis to add in the user-specified $optctl{RA}.

# In order to make it easier in the future to alter this part of the
# code if we change the format of the input file, the location of all
# of the coordinate values are given relative to the location of the
# *first* coordinate value (position 0).  In order to get to the
# actual array positions (which start with zero), we need to add in a
# column offset to take into account any columns which occur before
# the coordinate values.  This offset is $optctl{coord_elem_offset}.

# Note that the -w flag on perl will warn about many of these being 
# used only once.  That's not a problem.  

$ra_to_dec_pos_offset = 1;  # If we're working on an RA value and we want 
                            # the corresponding DEC value, how many
                            # columns do we shift over to get it?
$dec_to_ra_pos_offset = -1; # Similar to previous but for going from 
                            # a DEC value to the column with the 
                            # corresponding RA value. 

@circle_ra            = (0);  # For a circle, the only DEC-related
			      # coordinate is in array position 0 +
			      # $optctl{coord_elem_offset}.
@circle_dec           = (1);  # For a circle, the only RA-related
			      # coordinate is in array position 1 +
			      # $$optctl{coord_elem_offset}.

@ellipse_ra           = (0);  # See above for @circle_dec.
@ellipse_dec          = (1); 

@box_ra               = (0);        
@box_dec              = (1);      

@annulus_ra           = (0);
@annulus_dec          = (1);

@compass_ra           = (0);
@compass_dec          = (1);

@vector_ra            = (0);
@vector_dec           = (1);  

@line_ra              = (0,2); # For a line, there are two DEC-related
@line_dec             = (1,3); # coordinates and two RA-related
			       # coordinates.

@ruler_ra             = (0,2); 
@ruler_dec            = (1,3);

@projection_ra        = (0,2);
@projection_dec       = (1,3);

# That's enough for now.  If we want more element types, we can add
# them later.  Note that polygon is more complicated because it has an
# unspecified number of position values, depending on the number of
# sides.  If we ever want to use a polygon we'll have to add code in
# the section where we read in the position input file to deal with
# this.



# Set the following variables to reflect the choice 
# of degrees as the internal position unit.  
$internal_unit       = "deg"; # Used to help construct variable names 
                              # later in the code. 
$ra_out_unit_symbol  = 'd';
$dec_out_unit_symbol = 'd';

# If the following need to be altered, we have bigger problems than
# just this script.

$pi                  = 3.141592654;
$deg_to_deg          = 1.0;        # Needed later for auto 
                                   # detection and conversion 
                                   # of units. 

$radian_to_deg       = 360.0 / (2.0 * $pi); 
$arcmin_to_deg       = 1 / 60.0; 
$arcsec_to_arcmin    = 1 / 60.0; 
$arcsec_to_deg       = $arcsec_to_arcmin * $arcmin_to_deg; 

$deg_to_radian       = 1 / $radian_to_deg; 

# This converts the angle used internally within the code to 
# radians. 
$angle_to_radian     = $deg_to_radian;

 

#  ........... End of setting defaults.         ...............






#  .......... Read the fiber position file & generate the region file.   ..............


open IFU_ELEMENT_POSITION_FILE,   "<$ifu_element_position_file";
open OUTPUT_REGION_FILE,  ">$output_region_file";


# Call a subroutine to get the coordinate values onto a standard
# system that's easy to deal with in the code: decimal degrees.

foreach $coord_name ('RA','DEC') {
   $current_coord = $optctl{$coord_name}; 
   $coord_type    = ($coord_name =~ /ra/i) ? 'ra' : 'dec'; 

   # Name the coordinate variable used throughout the rest of 
   # the code in a manner that emphasizes it's the center of 
   # the IFU bundle.  
   $coord_variable_name = $coord_name . "_center"; 

   SWITCH:  {

     if ($current_coord =~ /^\s*([+-]?\d{1,2})[hHdD:\s]+?(\d{1,2})[Mm:\s]+?([\d\.]+)[sS\s]*$/) {
       # The coordinate is in sexagesimal, so convert to decimal. 
       ${$coord_variable_name} = sex_to_decimal ($1,$2,$3,$coord_type); 
       last SWITCH; 
     }

     if ($current_coord =~ /^\s*([+-\.\d]+)\s*$/) {
       # Already in decimal degrees.  Just set ${$coord_variable_name}
       # to the value we got from from the command line input
       # $optctl{$coord_name}.
       ${$coord_variable_name} = $current_coord; 
       # Do check to make sure we don't have a negative RA.
       warn "You have a negative value for the RA coordinate $coord_variable_name!" if 
           ($coord_type eq 'ra' && $current_coord < 0); 
       last SWITCH;
      }

      # We shouldn't get here. 
      warn "Something is wrong with the formatting of $coord_variable_name = $current_coord.\n"; 

   }  # End of SWITCH. 

}  # Done with making sure all coordinates are decimal.  

 



$composite_ra    = $RA_center;
$composite_dec   = $DEC_center;
$composite_angle = 0; # just for testing.

# Put in the defaults for a composite region.
if ($optctl{mk_composite})  {
 
  print OUTPUT_REGION_FILE "\# composite($composite_ra${ra_out_unit_symbol}$optctl{ds9_delimeter} ",
                           "$composite_dec${dec_out_unit_symbol}$optctl{ds9_delimeter} ",
                           "$composite_angle) \|\| color=$optctl{color} ",
                           "width=$optctl{line_width}\n";

  $region_line_termination = ' || '; # Indicates that a region is part
				     # of a composite.

} else {

  $region_line_termination = ' '; # Each individual region is NOT part
				  # of a composite.  This string,
				  # which is post-pended to each
				  # region line in the output file, is
				  # therefore a blank which doesn't do
				  # anything.

}


# Now step through the input file, fiber by fiber and write out the
# corresponding regions.  Note that there is *no* provision to check
# for and correct situations in which the RA goes negative or > 360
# deg, or for situations in which DEC goes out of its range of -90 deg
# to +90 deg.  This does not appear to be necessary, because ds9
# appeared to know what to do with such out-of-range values in a test
# of the procedure around sky position (00:00:00, 00:00:00).

$ifu_rel_pos_unit_not_found_ra  = 0; # Used to give an advisory after 
                                     # the while loop, if needed. 
$ifu_rel_pos_unit_not_found_dec = 0; 

IFU_ELEMENTS:  while (<IFU_ELEMENT_POSITION_FILE>) {

  next if /^\s*\#/;  # Skip comment lines. 
  next if /^\s*$/;   # Skip blank lines. 

  chomp;   # Works on $_ by default, which in this context is the 
           # input line from the fiber position file. 

  @inline_bits  = split /$optctl{input_delimeter}/;  # Works on $_ by default. 

  # Start out by copying the input line to an output line.  We'll
  # modify what's needed below, and everything else stays the same.
  @outline_bits = @inline_bits; 

  $fiber_id     =  $inline_bits[$optctl{inp_fiber_id_pos}];  

  # What kind of ds9 region element are we dealing with here?  
  $element_type = $inline_bits[$optctl{inp_elem_type_pos}]; 

  # The string giving the element type may have blank 
  # spaces around it from the input file.  This will 
  # screw us up below when we try to make a variable 
  # name out of this string.  The solution is to remove 
  # any blank spaces from the variable $element_type.  
  # We don't want to purge the entire @outline_bits of 
  # blank spaces because at some point we may want to 
  # include strings (potentially with embedded spaces) 
  # in the proto-ds9 region file.  

  $element_type =~ s/\s//g;

  # Which array gives the relative column positions for DEC and RA
  # offsets?  These arrays were defined early in the code: @circle_ra,
  # @circle_dec, @ellipse_ra, etc.
  $ra_position_array_name  = $element_type . "_ra"; 
  $dec_position_array_name = $element_type . "_dec"; 

  # Make sure we have created an array that's defined. 
  # Do this by checking so see there's at least one element 
  # in the array.  
  warn "ERROR!  \$ra_position_array_name = $ra_position_array_name " . 
      "is unknown or has zero elements!\n" unless $#{$ra_position_array_name} >= 0; 
  warn "ERROR!  \$dec_position_array_name = $dec_position_array_name " . 
      "is unknown or has zero elements!\n" unless $#{$dec_position_array_name} >= 0;




  # Now add the $DEC_center value specified by the user to all of the
  # DEC offsets given in the input line.  Step through the array giving
  # the relative column positions of the DEC coordinate values and add
  # the appropriate column offset to get the real position in
  # @inline_bits.

  # Need to do the DEC first because the RA needs a cos (DEC) term.  

  # We constructed the foreach loop in the following manner rather than e.g.
  #     foreach $dec_pos (...) 
  #     $dec_pos += $optctl{coord_elem_offset};
  # because if we had done it that way, the += would have incremented
  # the values in the @{$dec_position_array_name} and screwed
  # everything up!
  
  foreach $raw_pos (@{$dec_position_array_name}) {
    $dec_pos = $raw_pos + $optctl{coord_elem_offset};
    # Extract the numerical value and any strings indicating units
    # for each of the DEC coordinate values.  Test for success while
    # we're at it.  See page 81 in the Camel book for this
    # incantation.
    warn "Position $dec_pos does not appear to be correctly formated " .
         "for fiber number $fiber_id \n" unless 
         ( ($value,$unit_string) = $inline_bits[$dec_pos] =~ /([-+\d\.]+)([\s\D]*)/ );

    $_ = $unit_string; # Save typing in the switch below. 

    SWITCH: {
      $unit = "deg",    last SWITCH if /^\s*d/i; 
      $unit = "arcmin", last SWITCH if /^\s*(?:m|arcm|\')/i; 
      $unit = "arcsec", last SWITCH if /^\s*(?:s|arcs|\")/i; 
      $unit = "radian", last SWITCH if /^\s*r/i;
      $unit = $optctl{ifu_rel_pos_unit_dec};  # Default if nothing 
                                              # is specified. 
      $ifu_rel_pos_unit_not_found_dec++;      # Allows us to make an 
                                              # that the default 
                                              # unit is assumed. 
      }

    # Get the correct unit convesion factor. The values for 
    # the conversions are defined near the top of the script. 
    $unit_conversion_factor_string = $unit . "_to_" . $internal_unit; 
    $unit_conversion_factor        = ${$unit_conversion_factor_string}; 

    $dec_relative_value            = $value * $unit_conversion_factor; 
    # This is still relative to the center of the IFU bundle.   


    # $DEC_center is the coordinate supplied by the user.  $dec refers
    # to the coordinate of the specific region element we're working
    # on right now.
 
    $dec                           = $dec_relative_value + $DEC_center; 

    # Replace the DEC value in this array position in the output line 
    # with the newly calculated value. 
    $outline_bits[$dec_pos]        = "$dec$dec_out_unit_symbol"; 

  }  # End of the DEC foreach. 



  # Do the same thing for the RA values. See the note before the
  # foreach for DEC about why it's constructed in the manner it is.

  foreach $raw_pos (@{$ra_position_array_name}) {
      $ra_pos = $raw_pos + $optctl{coord_elem_offset}; 

      warn "Position $ra_pos does not appear to be correctly formated " .
           "for fiber number $fiber_id \n" unless 
           ( ($value,$unit_string) = $inline_bits[$ra_pos] =~ /([-+\d\.]+)([\s\D]*)/ );

      $_ = $unit_string; # Save typing in the switch below. 

      SWITCH: {
	  $unit = "deg",    last SWITCH if /^\s*d/i; 
          $unit = "arcmin", last SWITCH if /^\s*(?:m|arcm|\')/i; 
          $unit = "arcsec", last SWITCH if /^\s*(?:s|arcs|\")/i; 
          $unit = "radian", last SWITCH if /^\s*r/i;
          $unit = $optctl{ifu_rel_pos_unit_ra}; # Default if nothing 
                                                # is specified. 
          $ifu_rel_pos_unit_not_found_ra++;     # Allows us to make an 
                                                # that the default 
                                                # unit is assumed. 
      }

      $unit_conversion_factor_string = $unit . "_to_" . $internal_unit; 
      $unit_conversion_factor = ${$unit_conversion_factor_string}; 

      $ra_relative_value = $value * $unit_conversion_factor; 
 
      # Now add in the RA position of the center of the IFU bundle on
      # the sky.  Because $ra_relative_value is an offset on the sky
      # (actual angle moved on sky), its value must be divided by cos
      # (DEC) to get a *coordinate* offset which can be added to the
      # $RA_center coordinate corresponding to the center of the IFU.
     
      # Since the DEC values for this particular element have already
      # been processed, we can pull out the corresponding DEC value.
      # This DEC value will already be in the internal unit used by
      # the code, so we can use the $angle_to_radian conversion
      # factor.  We need to get the unit symbol stripped off of the
      # value of the DEC but otherwise we don't need that unit string.

      $dec_pos = $ra_pos + $ra_to_dec_pos_offset; 
      warn "Something wrong in getting the already processed " .
      "DEC value at array position \$dec_pos =  $dec_pos \n" .
      "in the foreach \$raw_pos loop. \n" unless 
      ( ($dec,$unit_string) = $outline_bits[$dec_pos] =~ /([-+\d\.]+)([\s\D]*)/ );

      $ra = $RA_center + $ra_relative_value/cos($dec * $angle_to_radian);

      $outline_bits[$ra_pos] = "$ra$ra_out_unit_symbol"; 
  }  # End of the RA foreach. 
  
  

  # Get rid of the fiber id column in @outline_bits, since ds9 doesn't
  # recognize this, but keep the relative order of everything else the
  # same.  Don't just assume the fiber id will always be the first
  # element (in which case we could have just used 'shift').

  for ($i = 0; $i <= $#outline_bits; $i++){
      push @outstring_bits, $outline_bits[$i] unless $i == $optctl{inp_fiber_id_pos}; 
  }


  # Now join everything together.  Use the specified ds9 delimeter to 
  # separate the various entries in the region file. Add a space for 
  # readability.  

  $outstring = join "$optctl{ds9_delimeter} ", @outstring_bits; 

  # We shouldn't have a delimeter after the string which specifies 
  # the element type, such as "circle" etc.  The easiest way to 
  # do this is to join everything and then take out the first 
  # occurence of the delimeter, since the region type is at the 
  # beginning of the line in the ds9 region file.  The easiest 
  # way to do this is a pattern match and replace in which we 
  # do *not* turn on global replacement. 

  $outstring =~ s/$optctl{ds9_delimeter}//; 

  # Since we used push to fill @outstring_bits, this array will 
  # retain all of the previous lines that we processed unless 
  # we clean it out each time though the while loop. 

  $#outstring_bits = -1; 

  print OUTPUT_REGION_FILE $outstring,$region_line_termination,"\n"; 


}

warn "Advisory:  At least one entry in the input file $ifu_element_position_file \n" . 
     "   does not have a unit listed for the relative RA offset values.  \n" . 
     "   The default unit of $optctl{ifu_rel_pos_unit_ra} was assumed. \n\n" if 
     ($ifu_rel_pos_unit_not_found_ra); 

warn "Advisory:  At least one entry in the input file $ifu_element_position_file \n" . 
     "   does not have a unit listed for the relative DEC offset values.  \n" . 
     "   The default unit of $optctl{ifu_rel_pos_unit_dec} was assumed. \n\n" if 
     ($ifu_rel_pos_unit_not_found_dec); 





# ..............  Clean up.   ................................

print STDOUT "Please find your new region file named   $output_region_file   in your ",
             "current (or otherwise specified) directory.\n",
             "Happy observing!\n" ;

close IFU_ELEMENT_POSITION_FILE; 
close OUTPUT_REGION_FILE;








#################    SUBROUTINES      #####################


sub sex_to_decimal {

    my ($deg_or_hr, $min, $sec, $type) = @_;
    $min                  += $sec/60.0;
    $deg_or_hr            += (($deg_or_hr =~ /-/) ? -$min/60.0 : $min/60.0);
    if ($type              =~ /^\s*ra?\s*/i) {                # We were sent an ra value.  
        return $deg_or_hr * 15.0;
    }
    elsif ($type           =~ /^\s*d(ec|eclination)?\s*/i) {  # We were sent a dec value.  
        return $deg_or_hr;
    }
    else  {
        print STDERR "You have given an inappropriate coordinate ".
               "type = $type to subroutine \'sex_to_decimal\'\n";
        return -100;
    }

}


###############################################################




########    REVISION HISTORY  ##########################

# -- October/November, 2007.  
# The basic functionality of this script was written into a
# quick-n-dirty script w/o any command-line options (all values
# hardwired) to overlay sparsepak on an image of G515 for a WIYN
# proposal.  It was adjusted to also overlay the proposed Hexpak
# fibers, since this observing proposal was written at about the same
# time as the NSF ATI proposal.



# -- 08 DEC 2007.  
# The script was resurrected to aid in Sparsepak setups for a WIYN
# observing run (Marsha Wolf).  We wanted to figure out how to align
# the fiber bundle to maximize the number of potential companions to
# AGN (the hosts are the main targets) which could be observed
# simultaneously with the quasar and host.  Since there were now
# several targets, the quick-n-dirty script was inefficient, so a more
# proper script with command-line interaction and documentation was
# produced.
#
# This is version 1.0. 



# -- 27 MAR 2009.  
# A crucial error was found in version 1.0.  $fiber_ra was not divided
# by cos($dec) before being added to $ra!  
# 
# This fix became version 1.1.



# -- 26 OCT 2009. 
# The script was modified to accomodate rectangular apertures. 
# The variable names related to the mask definition were given 
# less circle-centric names: 
# --ifu_element_shape    
# --ifu_element_unit   
# --ifu_element_radius  
# Two new parameters were added: 
# --ifu_element_width  
# --ifu_element_length 
# 
# The string written to OUTPUT_REGION_FILE was broken into 
# two pieces:
# 
# 1) $outstring_back is the backend of each line in the region file.
# It is defined first, before the IFU_ELEMENTS loop, via a case
# structure that currently accepts the strings "circle" or "box" for
# the region element shape.  (An entry of "rectangle" is converted to
# "box" earlier in the script.)
# 
# 2) $outstring_front is defined within the IFU_ELEMENTS loop.  It 
# contains the same front matter as in earlier versions. 
# 
# $outstring_front and $outstring_back are printed together on a 
# line in the output file at the end of the IFU_ELEMENTS loop. 
#
# 
# Since this is a major increase in flexibility, the version 
# is incremented to 2.0. 
# 
# The basename of the script was changed from mk_sparsepak_reg
# to mk_ifu_reg.  Hence, the progression of the script went from 
# mk_sparsepak_reg_1.1.pl to mk_ifu_reg_2.0.pl



# -- 27 OCT 2009. 
# Added a provision for having comment lines, which start with 
# the comment character '#' as the first non-blank character, 
# in the IFU element position file.  We're getting enough of these, 
# between Sparsepak, and the various configurations of Hexpak 
# and Saltislice, that it'd be nice to have some information 
# about each element definition file.  In addition, the code 
# now skips blank lines in the IFU element position file. 
# 
# The script is now mk_ifu_reg_2.1.pl 



# -- 02 JAN 2011. 
# Added the ability to enter coordinates for the fiber bundle 
# in a variety of formats.  Previously one could only enter 
# the coordinates in decimal degrees.  Now they can also be 
# entered in various sexagesimal formats, and these values 
# will be converted to decimal (used internally by the code) 
# on the fly.  The user does not have to specify the format, 
# as it will be auto detected as long as it's one of the 
# standard formats.  
# 
# The code to do this was copied, with a few modifications, 
# from the SALT RSS finder generator, mk_finder_2.1.1.pl, 
# including the subroutine sex_to_decimal which had in turn 
# came from earlier codes.  
#
# The script is now mk_ifu_reg_2.2.pl.  



# -- 02 JAN 2011. 
# The automatic creation of a composite region was not working 
# correctly before.  Two problems were identified:
# 1) The line width in the region file was set with 'line_width', 
#    when it should have been just 'width'.  This was fatal, as it 
#    prevented the region file from loading at all.  This was fixed. 
# 2) The composite region command at the beginning of the region 
#    file was not creating a composite region.  It turns out that 
#    every individual region in a composite must have its line 
#    terminated with the 'or' symbol ( || ).  This was fixed by 
#    simply post-pending a new string to the end of each output 
#    line that defined a region; if the region was meant to be 
#    composite, this string was || and otherwise it was a blank 
#    space.  
#
# Since this now appears to be working correctly, the default 
# for the mk_composite flag was changed from 0 (false, or don't 
# automatically make a composite region) to 1 (true, or *do* 
# automatically make a composite region).  
# 
# The script is now mk_ifu_reg_2.3.pl 



# -- 04 JAN 2011.  
# Fixed a bug in the new version which gave an error when the 
# coordinates were entered in decimal degrees.  The code was
# missing a line which put the decimal degree value into $RA_center 
# or $DEC_center as needed.  
# This line has been added:   ${$coord_name} = $current_coord; 
# 
# Added the line   $fiber_id    =  0;
# to keep perl from warning about the variable $fiber_id being used
# only once.  It doesn't change the performance of the script, just
# makes the execution cleaner.  Remove this if we ever start using
# $fiber_id for something.  The reason $fiber_id is there at all is to
# served as a place holder to capture the first column in the fiber
# position file.
# 
# The script is now mk_ifu_reg_2.3.1.pl 



# -- 04 JAN 2011 (later in the day).  
# Added commentary to STDOUT at the completion of the script, 
# at the request of a user. 
#
# Also did some minor cleanup of the documentation at the top 
# of the script. 
# 
# The script is now mk_ifu_reg_2.3.2.pl



# -- 23 FEB 2011.  

# Removed the command line options --RA_unit and --DEC_unit, including
# their listing and description in the comments at the beginning of
# the script plus their entry in the list of possible options and the
# setting of their defaults.  As part of this, all instances of
# $optctl{ra_unit} and $optctl{dec_unit} in the code were changed to
# the new variables $ra_unit and $dec_unit.  The default values of
# $ra_unit and $dec_unit are the same as before ('d'), but now they
# are set near the file opening commands at the top of the script.

# These changes have no effect on the operation of the code, but they
# will make it less likely that new users will become confused and
# inadvertently set these variables to incorrect values.  The units of
# RA and DEC originally could be set from the command line in
# anticipation of expanding the code to accomodate positions in
# formats other than decimal degrees.  However, when that capability
# was recently installed it was more sophisticated than originally
# planned and can autodetect the format and hence the units based on
# the user's entry.

# The variables $ra_unit and $dec_unit cannot be eliminated from the
# script, as these strings need to be entered into the ds9 region
# files.  The script uses decimal degrees internally, so $ra_unit and
# $dec_unit are hard-wired to reflect this.  This is not a loss of
# flexibility because if one wanted to alter $ra_unit and $dec_unit
# for some reason, one would also have to alter the subsequent code
# that assumes decimal degrees as the internal unit.

# Also updated the description of --ifu_element_radius at the top of
# the file to reflect the latest hexpak design as of this date.

# The script is now mk_ifu_reg_2.3.3.pl



# -- 24 FEB 2011. 

# We are starting substantial changes to the internal workings of the
# code.  The hope is that the external utilization of the code by the
# end user will stay the same and will in fact become simpler, as
# there should be fewer command line switches to worry about.

# Currently the input file contains only the (x,y) positions of the
# centers of each ds9 region element.  The type of element and any
# parameters are determined from command line switches.  This is easy
# for quick on-the-fly fiddling with potential IFU layouts, but as
# more shapes are required (e.g., rectangles for potential SALT image
# slicers) we require more and more command line switches to control
# this.

# With the advent of Hexpak and its variable sized fibers, the code is
# becoming too burdensome, especially for end users.  As it stands
# now, a user will have to run the code twice, once with the large
# fiber definition file and once with the small fiber definition file.
# Then they will have to load each of these output files into ds9 and
# make them composite.  We want to retain the ability of the user to
# run the code once and have the ds9 region file already be a complete
# composite.

# In order to retain end user simplicity, the code is being altered
# such that all of the potentially variable elements, such as
# position, shape, radius, etc., must be in the IFU element position
# input file.  The input file will become an almost complete ds9
# region file but with relative positions rather than absolute
# positions.  The job of the code becomes to simply change the
# relative positions to absolute and to add any other elements needed
# by the final ds9 region file, such as color, line width, etc.  This
# puts the burden of describing variable size and shape IFUs on the
# designers, who must provide the expanded proto-ds9 region file,
# rather than on the end users.

# To start with, we are pulling out the following command 
# line switches:
# --ifu_element_shape
# --ifu_element_unit
# --ifu_element_width 
# --ifu_element_length
# --ifu_element_radius 
# For reference, here are the radii for some of our IFU elements:
#                  Sparsepak                      4.687/2 = 2.34  arcsec
#                  Densepak                       2.81/2  = 1.41  arcsec
#                  Hexpak @ nasmyth port (large)  2.81/2  = 1.41  arcsec
#                  Hexpak @ nasmyth port (small)  0.94/2  = 0.47  arcsec

# --delimeter was changed to --ds9_delimeter and --input_delimeter was
# added to alert the code to what to expect for the column separators
# in the input file.

# Eliminated the section of the code which defined $outstring_back. 

# Added the definitions of arrays to locate coordinate elements,
# @circle_dec, etc.

# Almost completely rewrote the IFU_ELEMENTS while loop which is the
# heart of the code.  This can now handle the more complex but more
# flexible IFU relative position input files.

# Added and updated a lot of documentation within the code to reflect
# the changes and describe what's going on in some detail.



# -- 25 FEB 2011. 

# Squashed a few bugs introduced in all of yesterday's coding. 

# Since all of this is a major change in the inner workings of the
# code and of its external functionality, the version is upgraded to
# mk_ifu_reg_3.0.pl.



# -- 28 FEB 2011. 

# The new version isn't listing the units for (ra, dec) position 
# or for radius (arcsec) in the lines describing the individual 
# fiber elements.  Fixed that.  

# For the radii or other size values, those units must be 
# in the IFU fiber position file.  

# For the location parameters, there is now a default unit that can be
# set from the command line, or the units can be included.  These
# units are now auto-detected.  This involved substantial reworking of
# the IFU_ELEMENTS while loop.

# Cleaned up some of the new hard-wired values, even though we don't 
# expect to change them very often, by making them command-line 
# switches. 
# $input_fiber_id_pos became $optctl{inp_fiber_id_pos}. 
# $input_element_type_pos became $optctl{inp_elem_type_pos}.
# $coord_element_offset became $optctl{inp_elem_type_pos}.  

# Eliminated --ra_offset and --dec_offset command line switches, 
# because at this point there is no anticipated need for them.  

# Changed $RA and $DEC to $RA_center and $DEC_center in order to:
# 1) Make it more clear that these were the coordinates of the 
#    center of the IFU bundle; and 
# 2) Make it easier in case anyone ever wants to port this script 
#    to another language that doesn't distinguish case in 
#    variable names.  We had been using both $RA and $ra, 
#    and similarly for DEC, which are distint variables in perl. 
#    $ra and $dec are retained. 
# Altered the "foreach $coord_name ('RA','DEC') " loop to 
# accomodate this change in variable name.  

# Changed $deg_to_radians to $angle_to_radian in the heart of the 
# code and set   $angle_to_radian = $deg_to_radian  near the 
# top to make the code a little cleaner.  

# Changed $ra_unit and $dec_unit to $ra_out_unit_symbol and 
# $dec_out_unit_symbol to make the variable names more 
# explicitly express their function. 

# Tested the code with ds9 at the position (00:00:00, 00:00:00).  It
# seems to work.  In particular, ds9 seems to know how to hangle RA
# and DEC values that are out of range.  That's a relief so we don't
# have to fool with this in our code.  

# Cleaned up some of the documentation in the code.  

# The version is updated to 
# mk_ifu_reg_3.1.pl
